/* This file is part of the KDE project
   Copyright (C) 2002 Simon Hausmann <hausmann@kde.org>

   This library is free software; you can redistribute it and/or
   modify it under the terms of the GNU Library General Public
   License as published by the Free Software Foundation; either
   version 2 of the License, or (at your option) any later version.

   This library is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
   Library General Public License for more details.

   You should have received a copy of the GNU Library General Public License
   along with this library; see the file COPYING.LIB.  If not, write to
   the Free Software Foundation, Inc., 59 Temple Place - Suite 330,
   Boston, MA 02111-1307, USA.
*/

#include "message.h"
#include "global.h"
#include "messagefactory.h"
#include "markup.h"
#include "ctcp.h"

#include <klocale.h>

#include <assert.h>

// can't use 'using' here as using does not affect new declarations
namespace IRC
{

template <typename T, typename _SizeType> 
class BoundPointer
{
public:
    typedef T ValueType;
    typedef const T * ConstIterator;
    typedef _SizeType SizeType;
    typedef QValueList<ValueType> ValueList;

    BoundPointer( ConstIterator start, SizeType maxSteps,
                  const ValueList &optionalEndItems = ValueList() )
    {
        m_current = start;
        m_end = start + maxSteps;
        m_endItems = optionalEndItems;
    }

    void operator++() { ++m_current; }
    bool atEnd()
    { 
        if ( m_current >= m_end )
            return true;
        return m_endItems.findIndex( currentValue() ) != -1;
    }

    ConstIterator current() const { return m_current; }
    ValueType currentValue() const { return *current(); }

    ConstIterator end() const { return m_end; }

    SizeType stepsLeft() const { return m_end - m_current; }

private:
    ConstIterator m_current;
    ConstIterator m_end;
    ValueList m_endItems;
};

}

using namespace IRC;

Message::~Message()
{
}

void Message::setParameter( uint index, const QString &value )
{
    while ( m_parameters.count() <= index )
        m_parameters << QString::null;

    m_parameters[ index ] = value;
    m_validityChecked = false;
}

bool Message::isNumericType() const
{
    return m_type.length() == 3 &&
           m_type[ 0 ].isDigit() &&
           m_type[ 1 ].isDigit() &&
           m_type[ 2 ].isDigit();
}

uint Message::numericType() const
{
    assert( isNumericType() );
    return m_type.toUInt();
}

bool Message::isValid() const
{
    if ( m_validityChecked )
        return m_cachedValidity;

    m_cachedValidity = checkValidity();
    m_validityChecked = true;
    return m_cachedValidity;
}

bool Message::checkValidity() const
{
    if ( m_type.isEmpty() )
        return false;

    if ( !m_parameters.isEmpty() )
    {
        // only the last parameter may have a space or colon

        QStringList::ConstIterator it = m_parameters.begin();
        QStringList::ConstIterator last = m_parameters.end();
        last--;
        for (; it != last; ++it )
        {
            bool hasSpaceOrColon = (*it).find( ' ' ) != -1 || 
                                   (*it).find( ':' ) != -1;

            if ( hasSpaceOrColon )
                return false;
        }
    }

    return true;
}

QString Message::serialize() const
{
    QString result;

    if ( !isValid() )
        return result;

    result += serializePrefix();

    result += serializeType();

    result += serializeParameters();

    result += "\r\n";

    return result;
}

QString Message::parameter( uint index ) const
{
    if ( m_parameters.count() < index )
        return QString::null;

    return m_parameters[ index ];
}

QString Message::serializePrefix() const
{
    if ( m_prefix.isEmpty() )
        return QString::null;

    return ':' + m_prefix + ' ';
}

QString Message::serializeType() const
{
    return m_type;
}

QString Message::serializeParameters() const
{
    QString result;

    QStringList::ConstIterator it = m_parameters.begin();
    while ( it != m_parameters.end() )
    {
        QStringList::ConstIterator current = it;
        ++it;
        bool isLast = ( it == m_parameters.end() );
        result += serializeParameter( *current, isLast );
    }

    return result;
}

QString Message::serializeParameter( const QString &param, bool isLast ) 
{
    QString result( QChar( ' ' ) );

    bool hasSpaceOrColon = param.find( ' ' ) != -1 || 
                           param.find( ':' ) != -1;

    if ( isLast && hasSpaceOrColon )
        result += ':';

    return result + param;
}

Message Message::fromString( const QString &input )
{
    Message msg;

    // atEnd() will return true if these are encountered
    ParsingState::ValueList endItems;
    endItems << QChar( '\r' ) << QChar( '\n' );

    ParsingState state( input.unicode(), input.length(), endItems );

    if ( input.length() == 0 )
        return Message();

    if ( !msg.parsePrefix( state ) )
        return Message();

    if ( !msg.parseType( state ) )
        return Message();

    if ( !msg.parseParams( state ) )
        return Message();

    return msg;
}

bool Message::parsePrefix( ParsingState &state )
{
    if ( state.atEnd() )
        return false;

    if ( state.currentValue() != ':' ) // nothing to do
        return true;

    ++state;

    ParsingState::ConstIterator start = state.current();

    while ( !state.atEnd() && state.currentValue() != ' ' )
        ++state;

    ParsingState::SizeType len = state.current() - start;

    if ( len == 0 )
        return false;

    m_prefix = QString( start, len );

    parseSpace( state );

    return true;
}

bool Message::parseType( ParsingState &state )
{
    if ( state.atEnd() )
        return false;

    ParsingState::ConstIterator start = state.current();

    while ( !state.atEnd() && state.currentValue().isLetterOrNumber() )
        ++state;

    ParsingState::SizeType len = state.current() - start;

    if ( len == 0 )
        return false;

    m_type = QString( start, len );
    return true;
}

bool Message::parseParams( ParsingState &state )
{
    parseSpace( state );

    if ( state.atEnd() ) // spaces at the end are no fatal error.
        return true;

    if ( state.currentValue() == ':' )
    {
        ++state;
        if ( state.atEnd() )
            return true;

        ParsingState::ConstIterator start = state.current();

        while ( !state.atEnd() )
            ++state;

        ParsingState::SizeType len = state.current() - start;

        m_parameters << QString( start, len );
        return true;
    }

    ParsingState::ConstIterator start = state.current();

    while ( !state.atEnd() && !state.currentValue().isSpace() )
        ++state;

    uint len = state.current() - start;

    if ( len == 0 )
        return false;

    m_parameters << QString( start, len );

    if ( !state.atEnd() )
        return parseParams( state );

    return true;
}

void Message::parseSpace( ParsingState &state )
{
    while ( !state.atEnd() && state.currentValue() == ' ' )
        ++state;
}

QString PrivmsgMessage::text() const
{
    QString result;

    QString nick = sender().nick();
    if ( !nick.isEmpty() )
        result = "<" + nick + "> ";

    result += Markup::tagColorCodes( plainText() );

    return result;
}

void PrivmsgMessage::setAction( const QString &actionText )
{
    QString actionStr = QString::fromLatin1( "\1ACTION" );
    setPlainText( actionStr + " " + actionText + "\1" );
}

bool PrivmsgMessage::isCTCPMessage() const
{
    return CTCPMessage::isCTCPMessage( plainText() );
}

QString NickMessage::description() const
{
    Person who = sender();

    if ( who.isNull() )
	return QString::null;

    return i18n( "%1 is now known as %2" ).arg( who.nick() ).arg( nick() );
}

JoinMessage::ChannelInfoList JoinMessage::channelInfo() const
{
    const QStringList &params = parameters();

    if ( params.count() == 0 || 
         params.count() > 2 )
        return ChannelInfoList();

    ChannelInfoList result;

    QStringList chanlist = QStringList::split( ',', params[ 0 ] );

    QStringList keylist;
    if ( params.count() == 2 )
        keylist = QStringList::split( ',', params[ 1 ] ); 

    QStringList::ConstIterator chanIt = chanlist.begin();
    QStringList::ConstIterator keyIt = keylist.begin();
    for (; chanIt != chanlist.end(); ++chanIt )
    {
        ChannelInfo nfo( *chanIt );
        if ( keyIt != keylist.end() )
        {
            nfo.key = *keyIt;
            ++keyIt;
        }
        result << nfo;
    }

    return result;
}

void JoinMessage::setChannelInfo( const JoinMessage::ChannelInfoList &channels )
{
    QStringList chans;
    QStringList keys;

    ChannelInfoList::ConstIterator it = channels.begin();
    for (; it != channels.end(); ++it )
    {
	chans << (*it).channel;
	QString key = (*it).key;
	if ( !key.isEmpty() )
	    keys << key;
    }

    QStringList arguments;
    arguments << chans.join( "," );

    if ( !keys.isEmpty() )
	arguments << keys.join( "," );

    setParameters( arguments );
}

QString JoinMessage::description( const QString &channel ) const
{
    Person person = sender();
    QString who = QString::fromLatin1( "%1 (%3@%4)" ).arg( person.nick() ).
	          arg( person.user() ).arg( person.host() );

    return Markup::tagInformation( i18n( "%1 has joined channel %2" ).arg( who ).arg( channel ) );
}

QString PartMessage::description( const QString &channel ) const
{
    Person person = sender();
    QString who = QString::fromLatin1( "%1 (%3@%4)" ).arg( person.nick() ).
	          arg( person.user() ).arg( person.host() );

    QString reas = reason();
    if ( !reas.isEmpty() )
        return Markup::tagInformation( i18n( "%1 has left channel %2 (%3)" ).arg( who ).arg( channel ).arg( reas ) );

    return Markup::tagInformation( i18n( "%1 has left channel %2" ).arg( who ).arg( channel ) );
}

QString QuitMessage::description() const
{
    Person person = sender();

    return i18n( "Signoff: %1 (%2)" ).arg( person.nick() ).arg( quitMessage() );
}

namespace
{
    template <class T>
    QMemArray<T> makeArray( const T *data, uint count )
    {
        QMemArray<T> array( count );
        array.setRawData( data, count );
        QMemArray<T> copy = array;
        copy.detach();
        array.resetRawData( data, count );
        return copy;
    }

};

#define MAKE_REPLY_INFO( replyCodes, errorCodes ) \
    NumericReplyInfo( makeArray( replyCodes, sizeof( replyCodes ) / sizeof( replyCodes[ 0 ] ) ), \
                      makeArray( errorCodes, sizeof( errorCodes ) / sizeof( errorCodes[ 0 ] ) ) )

#define ERROR_INFO( errorCodes ) \
    NumericReplyInfo( NumericMessageTypeArray(), makeArray( errorCodes, sizeof( errorCodes ) / sizeof( errorCodes[ 0 ] ) ) );

#define REPLY_INFO( replyCodes ) \
    NumericReplyInfo( makeArray( replyCodes, sizeof( replyCodes ) / sizeof( replyCodes[ 0 ] ) ), NumericMessageTypeArray() );

Message::NumericReplyInfo PassMessage::replyInfo()
{
    static const uint errorCodes[] = { ERR_NEEDMOREPARAMS, ERR_ALREADYREGISTRED };
    return ERROR_INFO( errorCodes );
}

Message::NumericReplyInfo NickMessage::replyInfo()
{
    static const uint errorCodes[] = { ERR_NONICKNAMEGIVEN, ERR_ERRONEUSNICKNAME, ERR_NICKNAMEINUSE, ERR_NICKCOLLISION };
    return ERROR_INFO( errorCodes );
}

Message::NumericReplyInfo UserMessage::replyInfo()
{
    static const uint errorCodes[] = { ERR_NEEDMOREPARAMS, ERR_ALREADYREGISTRED };
    static const uint replyCodes[] = { RPL_WELCOME, RPL_YOURHOST, RPL_CREATED, RPL_MYINFO };
    return MAKE_REPLY_INFO( replyCodes, errorCodes );
}

Message::NumericReplyInfo ServerMessage::replyInfo()
{
    static const uint errorCodes[] = { ERR_ALREADYREGISTRED };
    return ERROR_INFO( errorCodes );
}

Message::NumericReplyInfo OperMessage::replyInfo()
{
    static const uint replyCodes[] = { RPL_YOUREOPER };
    static const uint errorCodes[] = { ERR_NEEDMOREPARAMS, ERR_NOOPERHOST, ERR_PASSWDMISMATCH };
    return MAKE_REPLY_INFO( replyCodes, errorCodes );
}

Message::NumericReplyInfo SquitMessage::replyInfo()
{
    static const uint errorCodes[] = { ERR_NOPRIVILEGES, ERR_NOSUCHSERVER };
    return ERROR_INFO( errorCodes );
}

Message::NumericReplyInfo JoinMessage::replyInfo()
{
    static const uint replyCodes[] = { RPL_NAMREPLY, RPL_ENDOFNAMES };
    //static const uint replyCodes[] = { RPL_TOPIC, RPL_NAMREPLY, RPL_ENDOFNAMES };
    static const uint errorCodes[] = { ERR_NEEDMOREPARAMS, ERR_BANNEDFROMCHAN,
                                       ERR_INVITEONLYCHAN, ERR_BADCHANNELKEY,
                                       ERR_CHANNELISFULL, ERR_BADCHANMASK,
                                       ERR_NOSUCHCHANNEL, ERR_TOOMANYCHANNELS };
    return MAKE_REPLY_INFO( replyCodes, errorCodes );
}

Message::NumericReplyInfo PartMessage::replyInfo()
{
    static const uint errorCodes[] = { ERR_NEEDMOREPARAMS, ERR_NOSUCHCHANNEL, ERR_NOTONCHANNEL };
    return ERROR_INFO( errorCodes );
}

Message::NumericReplyInfo TopicMessage::replyInfo()
{
    static const uint replyCodes[] = { RPL_NOTOPIC, RPL_TOPIC, RPL_TOPICWHOTIME };
    static const uint errorCodes[] = { ERR_NEEDMOREPARAMS, ERR_NOTONCHANNEL, ERR_CHANOPRIVSNEEDED };
    return MAKE_REPLY_INFO( replyCodes, errorCodes );
}

Message::NumericReplyInfo NamesMessage::replyInfo()
{
    static const uint replyCodes[] = { RPL_NAMREPLY, RPL_ENDOFNAMES };
    return REPLY_INFO( replyCodes );
}

Message::NumericReplyInfo ListMessage::replyInfo()
{
    static const uint replyCodes[] = { RPL_LISTSTART, RPL_LIST, RPL_LISTEND };
    static const uint errorCodes[] = { ERR_NOSUCHSERVER };
    return MAKE_REPLY_INFO( replyCodes, errorCodes );
}

Message::NumericReplyInfo InviteMessage::replyInfo()
{
    static const uint replyCodes[] = { RPL_INVITING, RPL_AWAY };
    static const uint errorCodes[] = { ERR_NEEDMOREPARAMS, ERR_NOSUCHNICK, ERR_NOTONCHANNEL, ERR_USERONCHANNEL, ERR_CHANOPRIVSNEEDED };
    return MAKE_REPLY_INFO( replyCodes, errorCodes );
}

Message::NumericReplyInfo KickMessage::replyInfo()
{
    static const uint errorCodes[] = { ERR_NEEDMOREPARAMS, ERR_NOSUCHCHANNEL, ERR_BADCHANMASK, ERR_CHANOPRIVSNEEDED, ERR_NOTONCHANNEL };
    return ERROR_INFO( errorCodes );
}

Message::NumericReplyInfo VersionMessage::replyInfo()
{
    static const uint replyCodes[] = { RPL_VERSION };
    static const uint errorCodes[] = { ERR_NOSUCHSERVER };
    return MAKE_REPLY_INFO( replyCodes, errorCodes );
}

Message::NumericReplyInfo StatsMessage::replyInfo()
{
    static const uint replyCodes[] = { RPL_STATSCLINE, RPL_STATSNLINE, RPL_STATSILINE, RPL_STATSKLINE, RPL_STATSQLINE,
                                       RPL_STATSLLINE, RPL_STATSLINKINFO, RPL_STATSUPTIME, RPL_STATSCOMMANDS,
                                       RPL_STATSOLINE, RPL_STATSHLINE, RPL_ENDOFSTATS };
    static const uint errorCodes[] = { ERR_NOSUCHSERVER };
    return MAKE_REPLY_INFO( replyCodes, errorCodes );
}

Message::NumericReplyInfo LinksMessage::replyInfo()
{
    static const uint replyCodes[] = { RPL_LINKS, RPL_ENDOFLINKS };
    static const uint errorCodes[] = { ERR_NOSUCHSERVER };
    return MAKE_REPLY_INFO( replyCodes, errorCodes );
}

Message::NumericReplyInfo TimeMessage::replyInfo()
{
    static const uint replyCodes[] = { RPL_TIME };
    static const uint errorCodes[] = { ERR_NOSUCHSERVER };
    return MAKE_REPLY_INFO( replyCodes, errorCodes );
}

Message::NumericReplyInfo ConnectMessage::replyInfo()
{
    static const uint errorCodes[] = { ERR_NOSUCHSERVER, ERR_NOPRIVILEGES, ERR_NEEDMOREPARAMS };
    return ERROR_INFO( errorCodes );
}

Message::NumericReplyInfo TraceMessage::replyInfo()
{
    static const uint replyCodes[] = { RPL_TRACELINK, RPL_TRACECONNECTING, RPL_TRACEUNKNOWN, RPL_TRACEUSER, RPL_TRACENEWTYPE,
                                       RPL_TRACECLASS, RPL_TRACEHANDSHAKE, RPL_TRACEOPERATOR };
    static const uint errorCodes[] = { ERR_NOSUCHSERVER };
    return MAKE_REPLY_INFO( replyCodes, errorCodes );
}

Message::NumericReplyInfo AdminMessage::replyInfo()
{
    static const uint replyCodes[] = { RPL_ADMINME, RPL_ADMINLOC1, RPL_ADMINLOC2, RPL_ADMINEMAIL };
    static const uint errorCodes[] = { ERR_NOSUCHSERVER };
    return MAKE_REPLY_INFO( replyCodes, errorCodes );
}

Message::NumericReplyInfo InfoMessage::replyInfo()
{
    static const uint replyCodes[] = { RPL_INFO, RPL_ENDOFINFO };
    static const uint errorCodes[] = { ERR_NOSUCHSERVER };
    return MAKE_REPLY_INFO( replyCodes, errorCodes );
}

Message::NumericReplyInfo PrivmsgMessage::replyInfo()
{
    static const uint replyCodes[] = { RPL_AWAY };
    static const uint errorCodes[] = { ERR_NORECIPIENT, ERR_CANNOTSENDTOCHAN, ERR_WILDTOPLEVEL, ERR_NOSUCHNICK, ERR_NOTEXTTOSEND,
                                       ERR_NOTOPLEVEL, ERR_TOOMANYTARGETS };
    return MAKE_REPLY_INFO( replyCodes, errorCodes );
}

Message::NumericReplyInfo WhoMessage::replyInfo()
{
    static const uint replyCodes[] = { RPL_WHOREPLY, RPL_ENDOFWHO };
    static const uint errorCodes[] = { ERR_NOSUCHSERVER };
    return MAKE_REPLY_INFO( replyCodes, errorCodes );
}

Message::NumericReplyInfo WhoisMessage::replyInfo()
{
    static const uint replyCodes[] = { RPL_WHOISUSER, RPL_WHOISCHANNELS, RPL_AWAY, RPL_WHOISIDLE, RPL_ENDOFWHOIS,
                                       RPL_WHOISSERVER, RPL_WHOISOPERATOR, RPL_WHOISIDENTIFIED };
    static const uint errorCodes[] = { ERR_NOSUCHSERVER, ERR_NONICKNAMEGIVEN, ERR_NOSUCHNICK };
    return MAKE_REPLY_INFO( replyCodes, errorCodes );
}

Message::NumericReplyInfo WhowasMessage::replyInfo()
{
    static const uint replyCodes[] = { RPL_WHOWASUSER, RPL_WHOISSERVER, RPL_ENDOFWHOWAS };
    static const uint errorCodes[] = { ERR_NONICKNAMEGIVEN, ERR_WASNOSUCHNICK };
    return MAKE_REPLY_INFO( replyCodes, errorCodes );
}

Message::NumericReplyInfo KillMessage::replyInfo()
{
    static const uint errorCodes[] = { ERR_NOPRIVILEGES, ERR_NEEDMOREPARAMS, ERR_NOSUCHNICK, ERR_CANTKILLSERVER };
    return ERROR_INFO( errorCodes );
}

Message::NumericReplyInfo PongMessage::replyInfo()
{
    static const uint errorCodes[] = { ERR_NOORIGIN, ERR_NOSUCHSERVER };
    return ERROR_INFO( errorCodes );
}

Message::NumericReplyInfo PingMessage::replyInfo()
{
    static const uint errorCodes[] = { ERR_NOORIGIN, ERR_NOSUCHSERVER };
    return ERROR_INFO( errorCodes );
}

Message::NumericReplyInfo AwayMessage::replyInfo()
{
    static const uint replyCodes[] = { RPL_UNAWAY, RPL_NOWAWAY };
    return REPLY_INFO( replyCodes );
}

Message::NumericReplyInfo RehashMessage::replyInfo()
{
    static const uint replyCodes[] = { RPL_REHASHING };
    static const uint errorCodes[] = { ERR_NOPRIVILEGES };
    return MAKE_REPLY_INFO( replyCodes, errorCodes );
}

Message::NumericReplyInfo RestartMessage::replyInfo()
{
    static const uint errorCodes[] = { ERR_NOPRIVILEGES };
    return ERROR_INFO( errorCodes );
}

Message::NumericReplyInfo SummonMessage::replyInfo()
{
    static const uint replyCodes[] = { RPL_SUMMONING };
    static const uint errorCodes[] = { ERR_NORECIPIENT, ERR_FILEERROR, ERR_NOLOGIN, ERR_NOSUCHSERVER };
    return MAKE_REPLY_INFO( replyCodes, errorCodes );
}

Message::NumericReplyInfo UsersMessage::replyInfo()
{
    static const uint replyCodes[] = { RPL_USERSSTART, RPL_USERS, RPL_NOUSERS, RPL_ENDOFUSERS };
    static const uint errorCodes[] = { ERR_NOSUCHSERVER, ERR_FILEERROR, ERR_USERSDISABLED };
    return MAKE_REPLY_INFO( replyCodes, errorCodes );
}

Message::NumericReplyInfo UserhostMessage::replyInfo()
{
    static const uint replyCodes[] = { RPL_USERHOST };
    static const uint errorCodes[] = { ERR_NEEDMOREPARAMS };
    return MAKE_REPLY_INFO( replyCodes, errorCodes );
}

Message::NumericReplyInfo IsonMessage::replyInfo()
{
    static const uint replyCodes[] = { RPL_ISON };
    static const uint errorCodes[] = { ERR_NEEDMOREPARAMS };
    return MAKE_REPLY_INFO( replyCodes, errorCodes );
}

/* vim: et sw=4
 */
